# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel

EVENT_CHOICES = {
    'convo.assigned': 'Conversation was assigned.',
    'convo.created': 'Conversation was created.',
    'convo.deleted': 'Conversation was deleted.',
    'convo.merged': 'Conversation was merged.',
    'convo.moved': 'Conversation was moved.',
    'convo.status': 'Conversation status was updated.',
    'convo.tags': 'Conversation tags were updated.',
    'convo.customer.reply.created': 'The customer replied to the '
                                    'conversation.',
    'convo.agent.reply.created': 'An agent replied to the conversation.',
    'convo.note.created': 'A note was added to the conversation.',
    'customer.created': 'A customer was created.',
    'satisfaction.ratings': 'A rating was received.',
}

event_type = properties.StringChoice(
    'The event type that this object represents.',
    required=True,
    choices=sorted(EVENT_CHOICES.keys()),
    descriptions=EVENT_CHOICES,
)


class WebHookEvent(BaseModel):

    event_type = event_type
    record = properties.Instance(
        'The parsed data record that was received in the request.',
        instance_class=BaseModel,
        required=True,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .api_common import ApiCommon

from ..apis.web_hook import WebHook


class TestApiWebHook(ApiCommon):

    def test_create(self):
        self._test_method(
            WebHook, 'create', [self.mock_record], '/hooks.json',
            None, 'post', False,
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .. import BaseApi

from ..request_paginator import RequestPaginator
from ..models.web_hook import WebHook


class WebHook(BaseApi):
    """This represents the ``WebHook`` Endpoint.

    The following aspects are implemented:

    * `Create Web Hook
      <http://developer.helpscout.net/help-desk-api/hooks/create/>`_
      (:func:`helpscout.apis.web_hook.WebHook.create`)
    """

    __object__ = WebHook

    @classmethod
    def create(cls, session, web_hook):
        """Create a web hook.

        Note that creating a new web hook will overwrite the web hook that is
        already configured for this company. There is also no way to
        programmatically determine if a web hook already exists for the
        company. This is a limitation of the HelpScout API and cannot be
        circumvented.

        Args:
            session (requests.sessions.Session): Authenticated session.
            web_hook (helpscout.models.WebHook): The web hook to be created.

        Returns:
            bool: ``True`` if the creation was a success. Errors otherwise.
        """
        cls(
            '/hooks.json',
            data=web_hook.to_api(),
            request_type=RequestPaginator.POST,
            session=session,
        )
        return True



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import unittest

from ..models.person import Person


class TestPerson(unittest.TestCase):

    def new_record(self, type):
        test_values = {
            'firstName': 'Boo',
            'lastName': 'Radley',
            'type': type,
        }
        return Person.from_api(**test_values)

    def setter_tester(self, customer_person_type, expected):
        person = self.new_record('customer')
        person.customer_person_type = customer_person_type
        self.assertEqual(person.type, expected)

    def test_customer_person_type_customer(self):
        """It should properly set customer_person_type to True when person
        type is 'customer'"""
        self.assertTrue(self.new_record('customer').customer_person_type)

    def test_customer_person_type_user(self):
        """It should properly set customer_person_type to False when person
        type is not 'customer'"""
        self.assertFalse(self.new_record('user').customer_person_type)

    def test_customer_person_type_setter_true(self):
        """It should properly set type to 'customer' when customer_person_type
        is set as True"""
        self.setter_tester(True, 'customer')

    def test_customer_person_type_setter_false(self):
        """It should properly set type to 'user' when customer_person_type
        is set as False"""
        self.setter_tester(False, 'user')



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties
import re

from .exceptions import HelpScoutValidationException


# Identify lowerCamelCase strings.
# https://stackoverflow.com/a/1176023/861399
REGEX_CAMEL_FIRST = re.compile(r'(.)([A-Z][a-z]+)')
REGEX_CAMEL_SECOND = re.compile(r'([a-z0-9])([A-Z])')


class BaseModel(properties.HasProperties):
    """This is the model that all other models inherit from.

    It provides some convenience functions, and the standard ``id`` property.
    """

    id = properties.Integer(
        'Unique identifier',
    )

    @classmethod
    def from_api(cls, **kwargs):
        """Create a new instance from API arguments.

        This will switch camelCase keys into snake_case for instantiation.

        It will also identify any ``Instance`` or ``List`` properties, and
        instantiate the proper objects using the values. The end result being
        a fully Objectified and Pythonified API response.

        Raises:
            HelpScoutValidationException: In the event that an unexpected
             property is received.

        Returns:
            BaseModel: Instantiated model using the API values.
        """

        vals = cls.get_non_empty_vals({
            cls._to_snake_case(k): v for k, v in kwargs.items()
        })
        vals.update({
            attr: cls._parse_property(attr, val) for attr, val in vals.items()
        })
        return cls(**vals)

    def get(self, key, default=None):
        """Return the field indicated by the key, if present."""
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def to_api(self):
        """Return a dictionary to send to the API.

        Returns:
            dict: Mapping representing this object that can be sent to the
             API.
        """
        vals = {}
        for attribute, attribute_type in self._props.items():
            prop = getattr(self, attribute)
            vals[self._to_camel_case(attribute)] = self._to_api_value(
                attribute_type, prop,
            )
        return vals

    def _to_api_value(self, attribute_type, value):
        """Return a parsed value for the API."""

        if not value:
            return None

        if isinstance(attribute_type, properties.Instance):
            return value.to_api()

        if isinstance(attribute_type, properties.List):
            return self._parse_api_value_list(value)

        return attribute_type.serialize(value)

    def _parse_api_value_list(self, values):
        """Return a list field compatible with the API."""
        try:
            return [v.to_api() for v in values]
        # Not models
        except AttributeError:
            return list(values)

    @staticmethod
    def get_non_empty_vals(mapping):
        """Return the mapping without any ``None`` values."""
        return {
            k: v for k, v in mapping.items() if v is not None
        }

    @classmethod
    def _parse_property(cls, name, value):
        """Parse a property received from the API into an internal object.

        Args:
            name (str): Name of the property on the object.
            value (mixed): The unparsed API value.

        Returns:
            mixed: A value compatible with the internal models.
        """

        prop = cls._props.get(name)

        if not prop:
            raise HelpScoutValidationException(
                '"%s" is not a valid property for "%s".' % (
                    name, cls,
                ),
            )

        if isinstance(prop, properties.Instance):
            return prop.instance_class.from_api(**value)

        elif isinstance(prop, properties.List):
            return cls._parse_property_list(prop, value)

        return value

    @staticmethod
    def _parse_property_list(prop, value):
        """Parse a list property and return a list of the results."""
        attributes = []
        for v in value:
            try:
                attributes.append(
                    prop.prop.instance_class.from_api(**v),
                )
            except AttributeError:
                attributes.append(v)
        return attributes

    @staticmethod
    def _to_snake_case(string):
        """Return a snake cased version of the input string.

        Args:
            string (str): A camel cased string.

        Returns:
            str: A snake cased string.
        """
        sub_string = r'\1_\2'
        string = REGEX_CAMEL_FIRST.sub(sub_string, string)
        return REGEX_CAMEL_SECOND.sub(sub_string, string).lower()

    @staticmethod
    def _to_camel_case(string):
        """Return a camel cased version of the input string.

        Args:
            string (str): A snake cased string.

        Returns:
            str: A camel cased string.
        """
        components = string.split('_')
        return '%s%s' % (
            components[0],
            ''.join(c.title() for c in components[1:]),
        )

    def __getitem__(self, item):
        """Return the field indicated by the key, if present.

        This is better than using ``getattr`` because it will not expose any
        properties that are not meant to be fields for the object.

        Raises:
            KeyError: In the event that the field doesn't exist.
        """
        self.__check_field(item)
        return getattr(self, item)

    def __setitem__(self, key, value):
        """Return the field indicated by the key, if present.

        This is better than using ``getattr`` because it will not expose any
        properties that are not meant to be fields for the object.

        Raises:
            KeyError: In the event that the field doesn't exist.
        """
        self.__check_field(key)
        return setattr(self, key, value)

    def __check_field(self, key):
        """Raises a KeyError if the field doesn't exist."""
        if not self._props.get(key):
            raise KeyError(
                'The field "%s" does not exist on "%s"' % (
                    key, self.__class__.__name__,
                ),
            )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .api_common import ApiCommon

from ..apis.conversations import Conversations

from ..models.search_conversation import SearchConversation
from ..models.thread import Thread


class TestApiConversations(ApiCommon):

    def test_create(self):
        data = {
            'conversation': self.mock_record.to_api(),
            'imported': True,
            'auto_reply': True,
            'reload': True,
        }
        args = [self.mock_record, True, True]
        self._test_method(
            Conversations, 'create', args, '/conversations.json',
            data, 'post', True,
        )

    def test_create_thread(self):
        thread = self.mock_record.thread
        data = {
            'thread': thread.to_api(),
            'imported': True,
            'reload': True,
        }
        args = [self.mock_record, self.mock_record.thread, True]
        uri = '/conversations/%d.json' % self.mock_record.id
        self._test_method(
            Conversations, 'create_thread', args, uri,
            data, 'post', True, Thread,
        )

    def test_delete(self):
        uri = '/conversations/%s.json' % self.mock_record.id
        self._test_method(
            Conversations, 'delete', [self.mock_record], uri,
            None, 'delete', False,
        )

    def test_find_customer(self):
        customer = self.mock_record.customer
        customer.id = 9876
        uri = '/mailboxes/%d/customers/%s/conversations.json' % (
            self.mock_record.id, customer.id,
        )
        args = [self.mock_record, customer]
        self._test_method(
            Conversations, 'find_customer', args, uri,
        )

    def test_find_user(self):
        user = self.mock_record.user
        user.id = 9876
        uri = '/mailboxes/%d/users/%s/conversations.json' % (
            self.mock_record.id, user.id,
        )
        args = [self.mock_record, user]
        self._test_method(
            Conversations, 'find_user', args, uri,
        )

    def test_get(self):
        uri = '/conversations/%s.json' % self.mock_record.id
        self._test_method(
            Conversations, 'get', [self.mock_record.id], uri,
            singleton=True,
        )

    def test_list(self):
        uri = '/mailboxes/%d/conversations.json' % self.mock_record.id
        self._test_method(
            Conversations, 'list', [self.mock_record], uri,
        )

    def test_list_folder(self):
        folder = self.mock_record.folder
        folder.id = 9876
        uri = '/mailboxes/%d/folders/%d/conversations.json' % (
            self.mock_record.id, folder.id,
        )
        self._test_method(
            Conversations, 'list_folder', [self.mock_record, folder], uri,
        )

    def test_search(self):
        data = {'query': "()"}
        self._test_method(
            Conversations, 'search', [[]], '/search/conversations.json', data,
            out_type=SearchConversation,
        )

    def test_update(self):
        uri = '/conversations/%s.json' % self.mock_record.id
        self._test_method(
            Conversations, 'update', [self.mock_record], uri, self.mock_record,
            singleton=True,
        )

    def test_update_thread(self):
        thread = self.mock_record.thread
        thread.id = 9876
        uri = '/conversations/%s/threads/%d.json' % (
            self.mock_record.id, thread.id,
        )
        self._test_method(
            Conversations, 'update_thread', [self.mock_record, thread], uri,
            thread, singleton=True, out_type=Thread,
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel
from .web_hook_event import event_type


class WebHook(BaseModel):

    url = properties.String(
        'The callback URL where Help Scout will post your webhook events. '
        'This is the script or location where you\'ll handle the data '
        'received from Help Scout.',
        required=True,
    )
    secret_key = properties.String(
        'A randomly-generated (by you) string of 40 characters or less used '
        'to create signatures for each webhook method. Help Scout uses this '
        'secret key to generate a signature for each webhook message. When '
        'the message is received at your callback URL, you can calculate a '
        'signature and compare to the one Help Scout sends. If the '
        'signatures match, you know it\'s from Help Scout.',
        required=True,
    )
    events = properties.List(
        'The events to subscribe to.',
        prop=event_type,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

__all__ = [
    'HelpScoutWebHook',
    'WebHookEvent',
]


import hmac
import json
import properties

from base64 import b64encode
from hashlib import sha1
from six import string_types

from ..models import Conversation, Customer, Rating, WebHook, WebHookEvent

from ..exceptions import HelpScoutSecurityException


class HelpScoutWebHook(WebHook):
    """This provides the ability to easily create & process web hook events.
    """

    helpscout = properties.Instance(
        'The authenticated HelpScout object. Used for create.',
        # This cannot be defined as a HelpScout object due to circular depends
        instance_class=object,
    )

    def create(self):
        """Create the web hook on HelpScout."""
        assert self.helpscout
        return self.helpscout.WebHook.create(self)

    def receive(self, event_type, signature, data_str):
        """Receive a web hook for the event and signature.

        Args:
            event_type (str): Name of the event that was received (from the
                request ``X-HelpScout-Event`` header).
            signature (str): The signature that was received, which serves as
                authentication (from the request ``X-HelpScout-Signature``
                header).
            data_str (str): The raw data that was posted by HelpScout
                to the web hook. This must be the raw string, because if it
                is parsed with JSON it will lose its ordering and not pass
                signature validation.

        Raises:
            helpscout.exceptions.HelpScoutSecurityException: If an invalid
                signature is provided, and ``raise_if_invalid`` is ``True``.

        Returns:
            helpscout.web_hook.WebHookEvent: The authenticated web hook
                request.
        """

        if not self.validate_signature(signature, data_str):
            raise HelpScoutSecurityException(
                'The signature provided by this request was invalid.',
            )

        return HelpScoutWebHookEvent(
            event_type=event_type,
            record=json.loads(data_str),
        )

    def validate_signature(self, signature, data, encoding='utf8'):
        """Validate the signature for the provided data.

        Args:
            signature (str or bytes or bytearray): Signature that was provided
                for the request.
            data (str or bytes or bytearray): Data string to validate against
                the signature.
            encoding (str, optional): If a string was provided for ``data`` or
                ``signature``, this is the character encoding.

        Returns:
            bool: Whether the signature is valid for the provided data.
        """

        if isinstance(data, string_types):
            data = bytearray(data, encoding)
        if isinstance(signature, string_types):
            signature = bytearray(signature, encoding)

        secret_key = bytearray(self.secret_key, 'utf8')
        hashed = hmac.new(secret_key, data, sha1)
        encoded = b64encode(hashed.digest())

        return encoded.strip() == signature.strip()


class HelpScoutWebHookEvent(WebHookEvent):
    """This represents an authenticated web hook request.

    Note that this object is meant to represent an authenticated Web Hook,
    and therefore is completely naive of all things authentication. Any
    request authentication/validation should happen in the ``WebHook``.
    """

    # Map the event prefixes to their corresponding data models.
    EVENT_PREFIX_TO_MODEL = {
        'convo': Conversation,
        'customer': Customer,
        'satisfaction': Rating,
    }

    def __init__(self, *args, **kwargs):
        """Parse raw record data if required.

        Args:
            record (dict or BaseModel): The record data that was received for
                the request. If it is a ``dict``, the data will be parsed
                using the proper model's ``from_api`` method.
        """
        if isinstance(kwargs.get('record'), dict):
            prefix, _ = kwargs['event_type'].split('.', 1)
            model = self.EVENT_PREFIX_TO_MODEL[prefix]
            kwargs['record'] = model.from_api(**kwargs['record'])
        super(WebHookEvent, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import mock
import unittest


class ApiCommon(unittest.TestCase):

    def setUp(self):
        super(ApiCommon, self).setUp()
        self.mock_record = mock.MagicMock()
        self.mock_class = mock.MagicMock()
        self.mock_session = mock.MagicMock()
        self.mock_record.id = 1234

    def _test_method(self, cls, method_name, method_args=None, uri=None,
                     data=None, request_type=None, singleton=False,
                     out_type=None):
        """Helper to test an API method call."""

        if method_args is None:
            method_args = []

        with mock.patch.object(cls, '__new__') as new:

            method = getattr(cls, method_name)
            method(self.mock_session, *method_args)

            new.assert_called_once()

            call_args = new.call_args[0]
            call_kwargs = new.call_args[1]

            self.assertEqual(call_kwargs['session'], self.mock_session)

            if uri is not None:
                self.assertEqual(call_args[1], uri)

            if data is not None:
                try:
                    self.assertEqual(call_kwargs['data'], data.to_api())
                # Simple object, not a model
                except AttributeError:
                    self.assertEqual(call_kwargs['data'], data)

            if request_type is not None:
                self.assertEqual(call_kwargs['request_type'], request_type)

            if singleton:
                self.assertTrue(call_kwargs['singleton'])

            if out_type is not None:
                self.assertEqual(call_kwargs['out_type'], out_type)
